from .advanced import Intersection, Projection
from .calculus import Derivate, Integrate
from .curves import Curve
from .functions import Function
from .knotspace import GeneratorKnotVector, KnotVector

__version__ = "1.0.8"

if __name__ == "__main__":
    pass
